/*
 * Waveform generator for STM32F4DISCOVERY
 *
 * PA0: User button (high on press)
 * PA1: Pulse length potentiometer
 * PA2: Low side offset potentiometer
 * PA3: High side offset potentiometer
 * PA4: Waveform output
 * PA6: Waveform output with less blowing up at 5V
 */

#include "stm32f4xx.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_syscfg.h"
#include "stm32f4xx_iwdg.h"
#include "stm32f4xx_dbgmcu.h"
#include "stm32f4xx_pwr.h"
#include "stm32f4xx_adc.h"
#include "system_stm32f4xx.h"
#include "config.h"

#include "serial.h"

#undef __STRICT_ANSI__
#include <math.h>

/* stk500v2 one-wire boot loader bits */
#define EXPECTED_CLOCK (HSE_VALUE * 21)
#define EXPECTED_KHZ (EXPECTED_CLOCK / 1000)
#define EXPECTED_MHZ (EXPECTED_CLOCK / 1000000)
#define QBIT_US 5
#define MAX_BIT_US 400

/* Waveform bits */
#define OUT_FREQ        5000            // Output waveform frequency
#define REAL_WAVE_RES   (21000/256)     // Waveform resolution
#define WAVE_RES        (21000/256)     // Waveform resolution
#define HALF_RES	(WAVE_RES / 2)
#define DAC_DHR12R1_ADDR 0x40007408     // DMA writes into this reg on every request
#define CNT_FREQ        42000000        // TIM6 counter clock (prescaled APB1)
#define TIM_PERIOD      ((CNT_FREQ)/((WAVE_RES)*(OUT_FREQ)))
#define ADC_SAMPLE_TIME ADC_SampleTime_28Cycles
//#define ADC_SAMPLE_TIME ADC_SampleTime_480Cycles

static uint16_t func[REAL_WAVE_RES];
static const char leader[] __attribute__((unused)) = "\377\377\177";

static void adc_init(void);
static void TIM6_Config(void);
static void DAC1_Config(void);

unsigned int delay_until(unsigned int ready_at)
{
	while (time_before(DWT->CYCCNT, ready_at))
		;
	return ready_at;
}

unsigned int delay_us(unsigned int us)
{
	int ready_at;

	ready_at = DWT->CYCCNT + us * EXPECTED_MHZ;
	while (time_before(DWT->CYCCNT, ready_at))
		;
	return ready_at;
}

unsigned int delay_ms(unsigned int ms)
{
	int ready_at;

	ready_at = DWT->CYCCNT + ms * EXPECTED_KHZ;
	while (time_before(DWT->CYCCNT, ready_at))
		;
	return ready_at;
}

int main()
{
	GPIO_InitTypeDef gpio;
	unsigned int i, decimator = 0;
	int bit;
	uint32_t pulse, low, high, avg, start, bit_time, nev, txwait;

	/* Allow debugger access during sleep (WFI). */
        DBGMCU_Config(DBGMCU_SLEEP, ENABLE);

        /* Set the NVIC priority group (bits of preemption vs priority). */
	NVIC_PriorityGroupConfig(NVIC_PriorityGroup_2);

        /* Enable DWT->CYCCNT for profiling and delaying. */
	if (!(CoreDebug->DEMCR & CoreDebug_DEMCR_TRCENA_Msk)) {
		CoreDebug->DEMCR|= CoreDebug_DEMCR_TRCENA_Msk;
		DWT->CTRL|= DWT_CTRL_CYCCNTENA_Msk;
	}

	/* Enable peripherals */
	RCC_AHB1PeriphClockCmd(
		RCC_AHB1Periph_GPIOA |
		RCC_AHB1Periph_GPIOB |	/* USART3 */
		RCC_AHB1Periph_GPIOC |
		RCC_AHB1Periph_GPIOD |	/* LEDs */
		RCC_AHB1Periph_DMA1, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_DAC, ENABLE);
	RCC_APB2PeriphClockCmd(
		RCC_APB2Periph_ADC1 |
		RCC_APB2Periph_ADC2 |
		RCC_APB2Periph_ADC3, ENABLE);

#if 0
	for (i = 0;i < WAVE_RES;i++)
		func[i] = sin(i * 2 * M_PI / WAVE_RES) * 2048 + 2047;
	for (i = 0;i < WAVE_RES;i++)
		func[i] = sin(i * 2 * M_PI / WAVE_RES) * 2048 + 2047;
#endif

	gpio.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_2 | GPIO_Pin_3 | GPIO_Pin_4;
	gpio.GPIO_Mode = GPIO_Mode_AN;
	gpio.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOA, &gpio);

	gpio.GPIO_Pin = GPIO_Pin_2;
//	gpio.GPIO_PuPd = GPIO_PuPd_DOWN;
	GPIO_Init(GPIOA, &gpio);

	gpio.GPIO_Pin = GPIO_Pin_3;
//	gpio.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &gpio);

	GPIO_StructInit(&gpio);
	gpio.GPIO_Pin = GPIO_Pin_1;
	gpio.GPIO_Mode = GPIO_Mode_OUT;
	gpio.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &gpio);
	GPIOC->BSRRL = GPIO_Pin_1;	/* Turn on PC1 as VDD */

	/* LED pins */
	gpio.GPIO_Pin = LED_PINS;
	GPIO_Init(LED_PORT, &gpio);

	serial_init();
	delay_ms(1);
#if 0
	/* Reset from various terminal corruption */
	serial_print("\r\033[K\033(B\033)B\017\033[m\r\n");

	/* Serial echo test */
	while (1) {
		if (serial_rx_ready())
			serial_putc(serial_read());
	}
#endif

#if 1
	/*
	 * Turnigy/Afro USB linker compatible signalling to/from SERIAL_UART.
	 */
#define GPIO_PWM (GPIO_Pin_4 | GPIO_Pin_6)
#define PWM_HIGH do { GPIOA->BSRRL = GPIO_PWM; } while (0)
#define PWM_LOW  do { GPIOA->BSRRH = GPIO_PWM; } while (0)
#define PWM_IN   (GPIOA->IDR & GPIO_Pin_6)
	gpio.GPIO_Pin = GPIO_PWM;
	gpio.GPIO_Mode = GPIO_Mode_IN;
	gpio.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &gpio);
	PWM_HIGH;	/* First action is high */
	txwait = 1;
	nev = DWT->CYCCNT + QBIT_US * EXPECTED_MHZ; /* First tx delay */

	while (1) {
	timeout:
		GREEN_OFF;
		if (txwait && time_after(DWT->CYCCNT, nev))
			txwait = 0;
		if (!txwait && PWM_IN && serial_rx_ready()) {
			const char *s = leader;
			char c;

			RED_ON;

			gpio.GPIO_Pin = GPIO_PWM;
			gpio.GPIO_Mode = GPIO_Mode_OUT;
			GPIO_Init(GPIOA, &gpio);
			nev = DWT->CYCCNT + QBIT_US * EXPECTED_MHZ;

			do {
				c = *s ?: serial_read();
				if (*s)
					s++;
				for (bit = 0;bit < 8;bit++) {
					PWM_HIGH;
					if (c & 1) {
						nev = delay_until(nev + QBIT_US * EXPECTED_MHZ) + QBIT_US * 2 * EXPECTED_MHZ;
					} else {
						nev = delay_until(nev) + QBIT_US * EXPECTED_MHZ;
						PWM_LOW;
						nev = delay_until(nev) + QBIT_US * EXPECTED_MHZ;
						PWM_HIGH;
						nev = delay_until(nev) + QBIT_US * EXPECTED_MHZ;
					}
					c>>= 1;
					PWM_LOW;
					nev = delay_until(nev) + QBIT_US * EXPECTED_MHZ;
				}
			} while (serial_rx_ready());
			PWM_HIGH;
			gpio.GPIO_Mode = GPIO_Mode_IN;
			txwait = 1;
			nev = delay_until(nev) + QBIT_US * EXPECTED_MHZ;
			GPIO_Init(GPIOA, &gpio);
			RED_OFF;
		}

		/* Loop while high */
		if (PWM_IN)
			continue;

		/*
		 * The leader is normally 23 1-bits followed by 1 0-bit.
		 * We just time between the first two high-going edges
		 * and wait for the first 0-bit. Using 7/8th of the bit
		 * time should allow us to catch all low-going edges in
		 * a bit transmission whlie stopping a little early to
		 * handle timing variances.
		 * ___
		 *    |___| = 1-bit
		 * _   _
		 *  |_| |_| = 0-bit
		 *
		 * Leader ignoring is done automatically by the 0-bit
		 * shifting through and triggering the serial send after
		 * starting "bit" with all 1-bits.
		 */

		bit_time = 0;
		bit = ~0;
		GREEN_ON;
		while (1) {
			if (!bit_time) {
				high = DWT->CYCCNT + MAX_BIT_US * EXPECTED_MHZ;
				while (!PWM_IN)
					if (time_after(DWT->CYCCNT, high))
						goto timeout;
				start = DWT->CYCCNT;
				high = start + MAX_BIT_US * EXPECTED_MHZ;
				while (PWM_IN)
					if (time_after(DWT->CYCCNT, high))
						goto timeout;
				while (!PWM_IN)
					if (time_after(DWT->CYCCNT, high))
						goto timeout;
				bit_time = DWT->CYCCNT - start;
			}

			start = DWT->CYCCNT;
			high = start + bit_time * 7 / 8;
			bit>>= 1;	/* Signed shift (ASR); top stays on */

			while (PWM_IN)
				if (time_after(DWT->CYCCNT, high))
					goto timeout;
			while (!PWM_IN)
				if (time_after(DWT->CYCCNT, high))
					goto onebit;
			bit&= ~0x100;
			while (PWM_IN)
				if (time_after(DWT->CYCCNT, high))
					goto timeout;
		onebit:
			if ((bit & 1) == 0) {
				serial_putc((bit >> 1) & 0xff);
				bit = ~0x100;
			}
			high = start + bit_time * 3 / 2;
			while (!PWM_IN)
				if (time_after(DWT->CYCCNT, high))
					goto timeout;
		}
	}
#endif

	DAC1_Config();
	TIM6_Config();
	adc_init();

	/*
	 * Wait for DMA half completion, then fill in 50% duty.
	 *
	 * We need to pay attention to the actual DMA cycle position and
	 * not fill during a time that would cause output jitter.
	 */
	mb();
	DMA1->HIFCR = DMA_HIFCR_CTCIF5 | DMA_HIFCR_CHTIF5; /* Clear TC and HT */
	while (!(DMA1->HISR & DMA_HISR_HTIF5))
		;
	mb();

	pulse = low = avg = 0;
	high = 4095;
	while (1) {
		pulse+= HALF_RES;
		for (i = 0;i < pulse;i++)
			func[i] = high;
		for (;i < REAL_WAVE_RES;i++)
			func[i] = low;

//		if (serial_rx_ready()) {
//			serial_read();
//		}
		if (!(decimator++ & 1023)) {
			mb();
//			serial_print(".");
//			serial_printf("pulse:%u low:%u high:%u\r\n",
//				(unsigned int)pulse, (unsigned int)low, (unsigned int)high);
		}

		pulse = low = high = avg = 0;
		while (!(DMA1->HISR & DMA_HISR_TCIF5)) {
			while (!(ADC1->SR & ADC_SR_EOC))
				;
			while (!(ADC2->SR & ADC_SR_EOC))
				;
			while (!(ADC3->SR & ADC_SR_EOC))
				;
			pulse+= ADC1->DR;
			low+=   ADC2->DR;
			high+=  ADC3->DR;
			avg+= 4096;
		}
		DMA1->HIFCR = DMA_HIFCR_CTCIF5 | DMA_HIFCR_CHTIF5; /* Clear TC and HT */
		pulse = (float)pulse * (float)HALF_RES / (float)avg;
		avg/= 4096;
		low/= avg;
		high/= avg;
	}
}

static void adc_init(void)
{
	ADC_InitTypeDef adc;
	ADC_CommonInitTypeDef acom;

	ADC_CommonStructInit(&acom);
	acom.ADC_Mode = ADC_Mode_Independent;
	acom.ADC_Prescaler = ADC_Prescaler_Div2;
	acom.ADC_DMAAccessMode = ADC_DMAAccessMode_Disabled;
	acom.ADC_TwoSamplingDelay = ADC_TwoSamplingDelay_20Cycles;
	ADC_CommonInit(&acom);

	ADC_StructInit(&adc);
	adc.ADC_ContinuousConvMode = ENABLE;
	ADC_Init(ADC1, &adc);
	ADC_Init(ADC2, &adc);
	ADC_Init(ADC3, &adc);
	ADC_RegularChannelConfig(ADC1, ADC_Channel_1, 1, ADC_SAMPLE_TIME);	/* ADC123_1 == PA1 */
	ADC_RegularChannelConfig(ADC2, ADC_Channel_2, 1, ADC_SAMPLE_TIME);	/* ADC123_2 == PA2 */
	ADC_RegularChannelConfig(ADC3, ADC_Channel_3, 1, ADC_SAMPLE_TIME);	/* ADC123_3 == PA3 */
	ADC_Cmd(ADC1, ENABLE);
	ADC_Cmd(ADC2, ENABLE);
	ADC_Cmd(ADC3, ENABLE);

	ADC_SoftwareStartConv(ADC1);
	ADC_SoftwareStartConv(ADC2);
	ADC_SoftwareStartConv(ADC3);
	while (!ADC_GetFlagStatus(ADC1, ADC_FLAG_EOC))
		;
}

static void TIM6_Config(void)
{
	TIM_TimeBaseInitTypeDef TIM6_TimeBase;

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6, ENABLE);

	TIM_TimeBaseStructInit(&TIM6_TimeBase);
	TIM6_TimeBase.TIM_Period        = 1;//(uint16_t)TIM_PERIOD;
	TIM6_TimeBase.TIM_Prescaler     = 0;
	TIM6_TimeBase.TIM_ClockDivision = 0;
	TIM6_TimeBase.TIM_CounterMode   = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM6, &TIM6_TimeBase);
	TIM_SelectOutputTrigger(TIM6, TIM_TRGOSource_Update);

	TIM_Cmd(TIM6, ENABLE);
}

static void DAC1_Config(void)
{
	DAC_InitTypeDef DAC_INIT;
	DMA_InitTypeDef DMA_INIT;

	DAC_StructInit(&DAC_INIT);
	DAC_INIT.DAC_Trigger        = DAC_Trigger_T6_TRGO;
	DAC_INIT.DAC_WaveGeneration = DAC_WaveGeneration_None;
	DAC_INIT.DAC_OutputBuffer   = DAC_OutputBuffer_Disable;
	DAC_Init(DAC_Channel_1, &DAC_INIT);

	DMA_DeInit(DMA1_Stream5);
	DMA_StructInit(&DMA_INIT);
	DMA_INIT.DMA_Channel            = DMA_Channel_7;
	DMA_INIT.DMA_PeripheralBaseAddr = (uint32_t)DAC_DHR12R1_ADDR;
	DMA_INIT.DMA_Memory0BaseAddr    = (uint32_t)&func;
	DMA_INIT.DMA_DIR                = DMA_DIR_MemoryToPeripheral;
	DMA_INIT.DMA_BufferSize         = REAL_WAVE_RES;
	DMA_INIT.DMA_PeripheralInc      = DMA_PeripheralInc_Disable;
	DMA_INIT.DMA_MemoryInc          = DMA_MemoryInc_Enable;
	DMA_INIT.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	DMA_INIT.DMA_MemoryDataSize     = DMA_MemoryDataSize_HalfWord;
	DMA_INIT.DMA_Mode               = DMA_Mode_Circular;
	DMA_INIT.DMA_Priority           = DMA_Priority_High;
	DMA_INIT.DMA_FIFOMode           = DMA_FIFOMode_Disable;
	DMA_INIT.DMA_FIFOThreshold      = DMA_FIFOThreshold_HalfFull;
	DMA_INIT.DMA_MemoryBurst        = DMA_MemoryBurst_Single;
	DMA_INIT.DMA_PeripheralBurst    = DMA_PeripheralBurst_Single;
	DMA_Init(DMA1_Stream5, &DMA_INIT);

	DMA_Cmd(DMA1_Stream5, ENABLE);
	DAC_Cmd(DAC_Channel_1, ENABLE);
	DAC_DMACmd(DAC_Channel_1, ENABLE);
}
