/*
 * serial.c : Serial communication.
 */

#include "serial.h"
#include "config.h"
#include "mini-printf.h"
#include "stm32f4xx_dma.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_rcc.h"
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>

struct serial_port p;
volatile int serial_tx_entry = 0;

unsigned char serial_read() {
	unsigned char c = p.rx_buf[SERIAL_BUFSIZE_RX - p.rx_ofs];

	if (!--p.rx_ofs)
		p.rx_ofs = SERIAL_BUFSIZE_RX;
	return c;
}

/*
 * tx_head is where we can start writing, up to before tx_tail.
 * tx_tail is where DMA will start or is ongoing.
 * tx_tail_dma is where tx_tail will be after DMA completes.
 */
static void serial_tx_dma_start() {
	unsigned int head = p.tx_head;
	unsigned int tail = p.tx_tail;

	mb();
	SERIAL_DMA_TX->M0AR = (uint32_t)&p.tx_buf[tail];
	if (head > tail) {
		ORANGE_ON;
		SERIAL_DMA_TX->NDTR = head - tail;
		p.tx_tail_dma = head;
	} else {
		SERIAL_DMA_TX->NDTR = SERIAL_BUFSIZE_TX - tail;
		p.tx_tail_dma = 0;
	}
	mb();
	SERIAL_DMA_TX->CR |= DMA_SxCR_EN | DMA_SxCR_TCIE;
}

/*
 * Fire off serial DMA if it is not already running. Since we might spin
 * on this if the buffer is full, reset the watchdog.
 */
static void serial_head_next(void) {
	p.tx_head = (p.tx_head + 1) % SERIAL_BUFSIZE_TX;
	if (!(SERIAL_DMA_TX->CR & DMA_SxCR_EN))
		serial_tx_dma_start();
	do {
		mb();
	} while (p.tx_head == p.tx_tail && (SERIAL_DMA_TX->CR & DMA_SxCR_EN));
}

void serial_putc(const char c) {
	serial_tx_entry = 1;
	mb();
	p.tx_buf[p.tx_head] = c;
	serial_head_next();
	mb();
	serial_tx_entry = 0;
}

void serial_write(const char *s, uint32_t l) {
	serial_tx_entry = 1;
	mb();
	while (l) {
		p.tx_buf[p.tx_head] = *s++;
		serial_head_next();
		l--;
	}
	mb();
	serial_tx_entry = 0;
}

void serial_print(const char *s) {
	serial_tx_entry = 1;
	mb();
	while (*s) {
		p.tx_buf[p.tx_head] = *s++;
		serial_head_next();
	}
	mb();
	serial_tx_entry = 0;
}

void serial_printvf(const char *fmt, va_list args) {
	char buf[128];
	int r = vsnprintf(buf, 128, fmt, args);
	if (r > 0 && r < 128)
		serial_write(buf, r);
}

void __attribute__((format(printf,1,2))) serial_printf(const char *fmt,...) {
	va_list args;

	va_start(args, fmt);
	serial_printvf(fmt, args);
	va_end(args);
}

void serial_flush(void) {
	while (p.tx_head != p.tx_tail)
		;
	while (!serial_tx_empty())
		;
}

void serial_init_core(uint32_t baud) {
	USART_InitTypeDef usart;

	/* Per stm32f3 reference: USART_BRR may only be written when UE=0 */
	SERIAL_UART->CR1&= ~USART_CR1_UE;
	usart.USART_BaudRate = baud;
	usart.USART_WordLength = USART_WordLength_8b;
	usart.USART_StopBits = USART_StopBits_1;
	usart.USART_Parity = USART_Parity_No;
	usart.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	usart.USART_HardwareFlowControl = SERIAL_FLOW_CONTROL;
	USART_Init(SERIAL_UART, &usart);
	USART_Cmd(SERIAL_UART, ENABLE);
}

void serial_init(void) {
	GPIO_InitTypeDef gpio;
	NVIC_InitTypeDef nvic;
	DMA_InitTypeDef dma;

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART3, ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOB, ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA1, ENABLE);

	/* Configure pins as alternate function push-pull */
	gpio.GPIO_Mode = GPIO_Mode_AF;
	gpio.GPIO_Speed = GPIO_Speed_50MHz;
	gpio.GPIO_OType = GPIO_OType_PP;
	gpio.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpio.GPIO_Pin = SERIAL_UART_TX_PIN;
	GPIO_Init(SERIAL_UART_PORT, &gpio);
	gpio.GPIO_PuPd = GPIO_PuPd_DOWN;
	gpio.GPIO_Pin = SERIAL_UART_RX_PIN;
	GPIO_Init(SERIAL_UART_PORT, &gpio);

	/* Enable alternate function for UART on TX, RX pins */
	GPIO_PinAFConfig(SERIAL_UART_PORT, SERIAL_UART_TX_PIN_SRC, GPIO_AF_USART3);
	GPIO_PinAFConfig(SERIAL_UART_PORT, SERIAL_UART_RX_PIN_SRC, GPIO_AF_USART3);

	/* Enable the TX interrupt */
	nvic.NVIC_IRQChannel = DMA1_Stream3_IRQn;
	nvic.NVIC_IRQChannelPreemptionPriority = 0;
	nvic.NVIC_IRQChannelSubPriority = 3;
	nvic.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvic);

	p.rx_head = p.rx_tail = 0;
	p.tx_head = p.tx_tail = 0;

	serial_init_core(SERIAL_BAUD_RATE);

	/* Configure DMA for rx */
	DMA_DeInit(SERIAL_DMA_RX);
	dma.DMA_Channel = SERIAL_DMA_CHANNEL_RX;
	dma.DMA_PeripheralBaseAddr = (uint32_t)&(SERIAL_UART->DR);
	dma.DMA_Memory0BaseAddr = (uint32_t)p.rx_buf;
	dma.DMA_DIR = DMA_DIR_PeripheralToMemory;
	dma.DMA_BufferSize = SERIAL_BUFSIZE_RX;
	dma.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dma.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dma.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
	dma.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
	dma.DMA_Mode = DMA_Mode_Circular;
	dma.DMA_Priority = DMA_Priority_Low;
	dma.DMA_FIFOMode = DMA_FIFOMode_Disable;
	dma.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
	dma.DMA_MemoryBurst = DMA_MemoryBurst_Single;
	dma.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
	DMA_Init(SERIAL_DMA_RX, &dma);
	USART_DMACmd(SERIAL_UART, USART_DMAReq_Rx, ENABLE);
	p.rx_ofs = DMA_GetCurrDataCounter(SERIAL_DMA_RX);
	/* Enable DMA Stream Half Transfer and Transfer Complete interrupt */
	//DMA_ITConfig(SERIAL_DMA_RX, DMA_IT_TC, ENABLE);
	//DMA_ITConfig(SERIAL_DMA_RX, DMA_IT_HT, ENABLE);
	DMA_Cmd(SERIAL_DMA_RX, ENABLE);

	/* Configure DMA for tx */
	DMA_DeInit(SERIAL_DMA_TX);
	dma.DMA_PeripheralBaseAddr = (uint32_t)&(SERIAL_UART->DR);
	dma.DMA_DIR = DMA_DIR_MemoryToPeripheral;
	dma.DMA_BufferSize = 0;
	dma.DMA_Mode = DMA_Mode_Normal;
	dma.DMA_FIFOMode = DMA_FIFOMode_Disable;
	DMA_Init(SERIAL_DMA_TX, &dma);
	USART_DMACmd(SERIAL_UART, USART_DMAReq_Tx, ENABLE);
	DMA_ITConfig(SERIAL_DMA_TX, DMA_IT_TC, ENABLE);
	/* Enable is done in serial_tx_dma_start() */
}

/* USART tx DMA IRQ */
void DMA1_Stream3_IRQHandler(void) {
	if (DMA_GetITStatus(SERIAL_DMA_TX, DMA_IT_TCIF3)) {
		DMA_ClearITPendingBit(SERIAL_DMA_TX, DMA_IT_TCIF3);
		DMA_Cmd(SERIAL_DMA_TX, DISABLE);
		p.tx_tail = p.tx_tail_dma;
		if (p.tx_head != p.tx_tail)
			serial_tx_dma_start();
	}
	ORANGE_OFF;
}
